package ru.dbotthepony.kstarbound.defs

import com.github.benmanes.caffeine.cache.Interner
import com.google.gson.Gson
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.io.json.builder.FactoryAdapter
import ru.dbotthepony.kstarbound.io.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.io.json.consumeNull
import ru.dbotthepony.kstarbound.util.ItemStack

/**
 * Прототип [ItemStack] в JSON файлах
 */
data class ItemReference(
	val item: Registry.Ref<IItemDefinition>,
	val count: Long = 1,
	val parameters: JsonObject = JsonObject()
) {
	init {
		require(item.key.isLeft) { "Can't reference item by ID" }
	}

	fun makeStack(): ItemStack {
		return ItemStack(item.entry ?: return ItemStack.EMPTY, count, parameters)
	}

	class Factory(val stringInterner: Interner<String> = Interner { it }) : TypeAdapterFactory {
		override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (type.rawType == ItemReference::class.java) {
				return object : TypeAdapter<ItemReference>() {
					private val regularObject = FactoryAdapter.createFor(ItemReference::class, JsonFactory(storesJson = false, asList = false), gson, stringInterner)
					private val regularList = FactoryAdapter.createFor(ItemReference::class, JsonFactory(storesJson = false, asList = true), gson, stringInterner)
					private val references = gson.getAdapter(TypeToken.getParameterized(Registry.Ref::class.java, IItemDefinition::class.java)) as TypeAdapter<Registry.Ref<IItemDefinition>>

					override fun write(out: JsonWriter, value: ItemReference?) {
						if (value == null)
							out.nullValue()
						else
							regularObject.write(out, value)
					}

					override fun read(`in`: JsonReader): ItemReference? {
						if (`in`.consumeNull())
							return null

						if (`in`.peek() == JsonToken.STRING) {
							return ItemReference(references.read(`in`))
						} else if (`in`.peek() == JsonToken.BEGIN_ARRAY) {
							return regularList.read(`in`)
						} else {
							return regularObject.read(`in`)
						}
					}
				} as TypeAdapter<T>
			}

			return null
		}
	}
}
